#ifndef LIO_LIVOX_ESTIMATOR_H
#define LIO_LIVOX_ESTIMATOR_H

#include <Eigen/Core>
#include <queue>
#include <iterator>
#include <future>
#include "ceresfunc.h"
#include "IMUIntegrator.h"
#include <chrono>
#include <fstream>
#include "LocalMapIvox.h"
#include "utils/TimerRecord.h"

class Estimator{
	typedef pcl::PointXYZINormal PointType;

    class Local_map
    {
     public:
        Local_map()
        {
            laserCloudCornerFromLocal.reset(new pcl::PointCloud<PointType>);
            laserCloudSurfFromLocal.reset(new pcl::PointCloud<PointType>);
            laserCloudNonFeatureFromLocal.reset(new pcl::PointCloud<PointType>);
            kdtreeCornerFromLocal.reset(new pcl::KdTreeFLANN<PointType>);
            kdtreeSurfFromLocal.reset(new pcl::KdTreeFLANN<PointType>);
            kdtreeNonFeatureFromLocal.reset(new pcl::KdTreeFLANN<PointType>);
          downSizeFilterCorner.setLeafSize(0.1, 0.1, 0.1);
          downSizeFilterSurf.setLeafSize(0.2, 0.2, 0.2);
          downSizeFilterNonFeature.setLeafSize(0.4, 0.4, 0.4);
        }
        void reset(pcl::PointCloud<PointType>::Ptr corner,pcl::PointCloud<PointType>::Ptr surf,
                   pcl::PointCloud<PointType>::Ptr nonf)
        {
            laserCloudCornerFromLocal=corner;
            laserCloudSurfFromLocal=surf;
            laserCloudNonFeatureFromLocal=nonf;
            kdtreeCornerFromLocal->setInputCloud(corner);
            kdtreeSurfFromLocal->setInputCloud(surf);
            kdtreeNonFeatureFromLocal->setInputCloud(nonf);
        }
        pcl::PointCloud<PointType>::Ptr full_cloud()
        {
            pcl::PointCloud<PointType>::Ptr cloud(new pcl::PointCloud<PointType>);
            *cloud+=(*laserCloudCornerFromLocal);
            *cloud+=(*laserCloudSurfFromLocal);
            *cloud+=(*laserCloudNonFeatureFromLocal);
            return cloud;
        }
        static int read_txt_cloud(pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud,std::string txt)
        {
            std::ifstream in(txt.c_str(), std::ios::in);
            if (!in)
            {
                return 0;
            }
            double x_ = 0, y_ = 0, z_ = 0;
            int r=0,g=0,b=0;
            int i = 0;
            while (!in.eof())
            {
                in >> x_ >> y_ >> z_>>r>>g>>b;
                if (in.fail())
                {
                    break;
                }
                pcl::PointXYZRGB Pt;
                Pt.x = x_;
                Pt.y = y_;
                Pt.z = z_;
                Pt.r= r;
                Pt.g=g;
                Pt.b=b;
                cloud->points.push_back(Pt);
                ++i;
            }
            in.close();
            //std::cout << "Import finished ... ..." << std::endl;
            return true;
        }

        bool load_map(std::string cloud_file)
        {
          pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZRGB>);

          read_txt_cloud(cloud, cloud_file);
          pcl::PointCloud<PointType>::Ptr Corner(new pcl::PointCloud<PointType>);
          pcl::PointCloud<PointType>::Ptr Surf(new pcl::PointCloud<PointType>);
          pcl::PointCloud<PointType>::Ptr Nonf(new pcl::PointCloud<PointType>);


          for (int i = 0; i < cloud->size(); ++i)
          {
            pcl::PointXYZRGB pt = cloud->points[i];
            PointType pt_temp;
            pt_temp.x = pt.x;
            pt_temp.y = pt.y;
            pt_temp.z = pt.z;
            if (pt.r == 255)
            {
              Corner->push_back(pt_temp);
            }
            else if (pt.g == 255)
            {
              Surf->push_back(pt_temp);
            }
            else
            {
              Nonf->push_back(pt_temp);
            }
          }

          downSizeFilterCorner.setInputCloud(Corner);
          downSizeFilterCorner.filter(*laserCloudCornerFromLocal);

          downSizeFilterSurf.setInputCloud(Surf);
          downSizeFilterSurf.filter(*laserCloudSurfFromLocal);

          downSizeFilterNonFeature.setInputCloud(Nonf);
          downSizeFilterNonFeature.filter(*laserCloudNonFeatureFromLocal);
          printf(" load map   corner:%d  surf:%d  nonf:%d\n", laserCloudCornerFromLocal->size(),
                 laserCloudSurfFromLocal->size(), laserCloudNonFeatureFromLocal->size());
          kdtreeCornerFromLocal->setInputCloud(laserCloudCornerFromLocal);
          kdtreeSurfFromLocal->setInputCloud(laserCloudSurfFromLocal);
          kdtreeNonFeatureFromLocal->setInputCloud(laserCloudNonFeatureFromLocal);
          return true;
        }
        /*pcl::PointCloud<PointType>::Ptr& corner_cloud(){return laserCloudCornerFromLocal;}
        pcl::PointCloud<PointType>::Ptr& surf_cloud(){return laserCloudSurfFromLocal;}
        pcl::PointCloud<PointType>::Ptr& nonf_cloud(){return laserCloudNonFeatureFromLocal;}*/

     public:
        pcl::VoxelGrid<PointType> downSizeFilterCorner;
        pcl::VoxelGrid<PointType> downSizeFilterSurf;
        pcl::VoxelGrid<PointType> downSizeFilterNonFeature;

        pcl::KdTreeFLANN<PointType>::Ptr kdtreeCornerFromLocal;
        pcl::KdTreeFLANN<PointType>::Ptr kdtreeSurfFromLocal;
        pcl::KdTreeFLANN<PointType>::Ptr kdtreeNonFeatureFromLocal;

        pcl::PointCloud<PointType>::Ptr laserCloudCornerFromLocal;
        pcl::PointCloud<PointType>::Ptr laserCloudSurfFromLocal;
        pcl::PointCloud<PointType>::Ptr laserCloudNonFeatureFromLocal;
    };

public:
	/** \brief slide window size */
	static const int SLIDEWINDOWSIZE = 2;

	/** \brief lidar frame struct */
	struct LidarFrame{
		pcl::PointCloud<PointType>::Ptr laserCloud;
		IMUIntegrator imuIntegrator;
		Eigen::Vector3d P;
		Eigen::Vector3d V;
		Eigen::Quaterniond Q;
		Eigen::Vector3d bg;
		Eigen::Vector3d ba;
		double timeStamp;
		LidarFrame(){
			P.setZero();
			V.setZero();
			Q.setIdentity();
			bg.setZero();
			ba.setZero();
			timeStamp = 0;
		}
	};

	/** \brief point to line feature */
	struct FeatureLine{
		Eigen::Vector3d pointOri;
		Eigen::Vector3d lineP1;
		Eigen::Vector3d lineP2;
		double error;
		bool valid;
		FeatureLine(Eigen::Vector3d  po, Eigen::Vector3d  p1, Eigen::Vector3d  p2)
						:pointOri(std::move(po)), lineP1(std::move(p1)), lineP2(std::move(p2)){
			valid = false;
			error = 0;
		}
		double ComputeError(const Eigen::Matrix4d& pose){
			Eigen::Vector3d P_to_Map = pose.topLeftCorner(3,3) * pointOri + pose.topRightCorner(3,1);
			double l12 = std::sqrt((lineP1(0) - lineP2(0))*(lineP1(0) - lineP2(0)) + (lineP1(1) - lineP2(1))*
																						(lineP1(1) - lineP2(1)) + (lineP1(2) - lineP2(2))*(lineP1(2) - lineP2(2)));
			double a012 = std::sqrt(
							((P_to_Map(0) - lineP1(0)) * (P_to_Map(1) - lineP2(1)) - (P_to_Map(0) - lineP2(0)) * (P_to_Map(1) - lineP1(1)))
							* ((P_to_Map(0) - lineP1(0)) * (P_to_Map(1) - lineP2(1)) - (P_to_Map(0) - lineP2(0)) * (P_to_Map(1) - lineP1(1)))
							+ ((P_to_Map(0) - lineP1(0)) * (P_to_Map(2) - lineP2(2)) - (P_to_Map(0) - lineP2(0)) * (P_to_Map(2) - lineP1(2)))
								* ((P_to_Map(0) - lineP1(0)) * (P_to_Map(2) - lineP2(2)) - (P_to_Map(0) - lineP2(0)) * (P_to_Map(2) - lineP1(2)))
							+ ((P_to_Map(1) - lineP1(1)) * (P_to_Map(2) - lineP2(2)) - (P_to_Map(1) - lineP2(1)) * (P_to_Map(2) - lineP1(2)))
								* ((P_to_Map(1) - lineP1(1)) * (P_to_Map(2) - lineP2(2)) - (P_to_Map(1) - lineP2(1)) * (P_to_Map(2) - lineP1(2))));
			error = a012 / l12;
			return error;
		}
	};

	/** \brief point to plan feature */
	struct FeaturePlan{
		Eigen::Vector3d pointOri;
		double pa;
		double pb;
		double pc;
		double pd;
		double error;
		bool valid;
		FeaturePlan(const Eigen::Vector3d& po, const double& pa_, const double& pb_, const double& pc_, const double& pd_)
						:pointOri(po), pa(pa_), pb(pb_), pc(pc_), pd(pd_){
			valid = false;
			error = 0;
		}
		double ComputeError(const Eigen::Matrix4d& pose){
			Eigen::Vector3d P_to_Map = pose.topLeftCorner(3,3) * pointOri + pose.topRightCorner(3,1);
			error = pa * P_to_Map(0) + pb * P_to_Map(1) + pc * P_to_Map(2) + pd;
			return error;
		}
	};

	/** \brief point to plan feature */
	struct FeaturePlanVec{
		Eigen::Vector3d pointOri;
		Eigen::Vector3d pointProj;
		Eigen::Matrix3d sqrt_info;
		double error;
		bool valid;
		FeaturePlanVec(const Eigen::Vector3d& po, const Eigen::Vector3d& p_proj, Eigen::Matrix3d sqrt_info_)
						:pointOri(po), pointProj(p_proj), sqrt_info(sqrt_info_) {
			valid = false;
			error = 0;
		}
		double ComputeError(const Eigen::Matrix4d& pose){
			Eigen::Vector3d P_to_Map = pose.topLeftCorner(3,3) * pointOri + pose.topRightCorner(3,1);
			error = (P_to_Map - pointProj).norm();
			return error;
		}
	};

	/** \brief non feature */
	struct FeatureNon{
		Eigen::Vector3d pointOri;
		double pa;
		double pb;
		double pc;
		double pd;
		double error;
		bool valid;
		FeatureNon(const Eigen::Vector3d& po, const double& pa_, const double& pb_, const double& pc_, const double& pd_)
						:pointOri(po), pa(pa_), pb(pb_), pc(pc_), pd(pd_){
			valid = false;
			error = 0;
		}
		double ComputeError(const Eigen::Matrix4d& pose){
			Eigen::Vector3d P_to_Map = pose.topLeftCorner(3,3) * pointOri + pose.topRightCorner(3,1);
			error = pa * P_to_Map(0) + pb * P_to_Map(1) + pc * P_to_Map(2) + pd;
			return error;
		}
	};

public:
	/** \brief constructor of Estimator
	*/
	Estimator(const float& filter_corner, const float& filter_surf,int ivox_nearby_type,int max_ivox_node,float resolution);

	~Estimator();

	void print();


	/** \brief construct sharp feature Ceres Costfunctions
	* \param[in] edges: store costfunctions
	* \param[in] m4d: lidar pose, represented by matrix 4X4
	*/
	void processPointToLine(std::vector<ceres::CostFunction *>& edges,
							std::vector<FeatureLine>& vLineFeatures,
							const pcl::PointCloud<PointType>::Ptr& laserCloudCorner,
							const Eigen::Matrix4d& exTlb,
							const Eigen::Matrix4d& m4d);

	void processPointToPlanVec(std::vector<ceres::CostFunction *>& edges,
							   std::vector<FeaturePlanVec>& vPlanFeatures,
							   const pcl::PointCloud<PointType>::Ptr& laserCloudSurf,
							   const Eigen::Matrix4d& exTlb,
							   const Eigen::Matrix4d& m4d);
				
	void processNonFeatureICP(std::vector<ceres::CostFunction *>& edges,
							  std::vector<FeatureNon>& vNonFeatures,
							  const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeature,
							  const Eigen::Matrix4d& exTlb,
							  const Eigen::Matrix4d& m4d);

	/** \brief Transform Lidar Pose in slidewindow to double array
		* \param[in] lidarFrameList: Lidar Poses in slidewindow
		*/
	void vector2double(const std::list<LidarFrame>& lidarFrameList);

	/** \brief Transform double array to Lidar Pose in slidewindow
		* \param[in] lidarFrameList: Lidar Poses in slidewindow
		*/
	void double2vector(std::list<LidarFrame>& lidarFrameList);

	/** \brief estimate lidar pose by matching current lidar cloud with map cloud and tightly coupled IMU message
		* \param[in] lidarFrameList: multi-frames of lidar cloud and lidar pose
		* \param[in] exTlb: extrinsic matrix between lidar and IMU
		* \param[in] gravity: gravity vector
		*/
	void EstimateLidarPose(std::list<LidarFrame>& lidarFrameList,
						   const Eigen::Matrix4d& exTlb,
						   const Eigen::Vector3d& gravity
						   /*nav_msgs::Odometry& debugInfo*/);

	void Estimate(std::list<LidarFrame>& lidarFrameList,
				  const Eigen::Matrix4d& exTlb,
				  const Eigen::Vector3d& gravity);

	pcl::PointCloud<PointType>::Ptr get_corner_map(){
		return m_global_map.laserCloudCornerFromLocal;
	}
	pcl::PointCloud<PointType>::Ptr get_surf_map(){
		return m_global_map.laserCloudSurfFromLocal;
	}
	pcl::PointCloud<PointType>::Ptr get_nonfeature_map(){
		return m_global_map.laserCloudNonFeatureFromLocal;
	}
	void MapIncrementLocal(const pcl::PointCloud<PointType>::Ptr& laserCloudCornerStack,
						   const pcl::PointCloud<PointType>::Ptr& laserCloudSurfStack,
						   const pcl::PointCloud<PointType>::Ptr& laserCloudNonFeatureStack,
						   const Eigen::Matrix4d& transformTobeMapped);


	/** \brief store map points */
    Local_map m_global_map;
    //LocalMapIvox local_map_;

    /*pcl::PointCloud<PointType> local_corner_cloud_;
    pcl::PointCloud<PointType> local_surf_cloud_;
    pcl::PointCloud<PointType> local_nonf_cloud_;*/
    double final_cost_=0;

 private:
	double para_PR[SLIDEWINDOWSIZE][6];
	double para_VBias[SLIDEWINDOWSIZE][9];
	MarginalizationInfo *last_marginalization_info = nullptr;
	std::vector<double *> last_marginalization_parameter_blocks;
	std::vector<pcl::PointCloud<PointType>::Ptr> laserCloudCornerLast;
	std::vector<pcl::PointCloud<PointType>::Ptr> laserCloudSurfLast;
	std::vector<pcl::PointCloud<PointType>::Ptr> laserCloudNonFeatureLast;




	pcl::PointCloud<PointType>::Ptr laserCloudCornerForMap;
	pcl::PointCloud<PointType>::Ptr laserCloudSurfForMap;
	pcl::PointCloud<PointType>::Ptr laserCloudNonFeatureForMap;
	Eigen::Matrix4d transformForMap;
	std::vector<pcl::PointCloud<PointType>::Ptr> laserCloudCornerStack;
	std::vector<pcl::PointCloud<PointType>::Ptr> laserCloudSurfStack;
	std::vector<pcl::PointCloud<PointType>::Ptr> laserCloudNonFeatureStack;

	pcl::VoxelGrid<PointType> downSizeFilterCorner;
	pcl::VoxelGrid<PointType> downSizeFilterSurf;
	pcl::VoxelGrid<PointType> downSizeFilterNonFeature;
	std::mutex mtx_Map;

	int localMapID = 0;

	double plan_weight_tan = 0.0;
	double thres_dist = 0.5;
};

#endif //LIO_LIVOX_ESTIMATOR_H
