//
// The MIT License (MIT)
//
// Copyright (c) 2019 Livox. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include "lddc.h"

#include <inttypes.h>
#include <math.h>
#include <stdint.h>

#include "lds_lidar.h"
#include "lds_lvx.h"

namespace livox_ros {

/** Lidar Data Distribute Control--------------------------------------------*/
Lddc::Lddc(double frq)
    :publish_frq_(frq){
  exit_=true;
  publish_period_ns_ = kNsPerSecond / publish_frq_;
  lds_ = nullptr;
  private_pub_= nullptr;
  private_imu_pub_=nullptr;
  capture_=nullptr;


};

Lddc::~Lddc() {

  if(capture_!= nullptr)
  {
    exit_=true;
    capture_->join();
    delete capture_;
    capture_= nullptr;
  }
  if (lds_) {
    lds_->PrepareExit();
  }

}

int32_t Lddc::GetPublishStartTime(LidarDevice *lidar, LidarDataQueue *queue,
                                  uint64_t *start_time,
                                  StoragePacket *storage_packet) {
  QueuePrePop(queue, storage_packet);
  uint64_t timestamp =
      GetStoragePacketTimestamp(storage_packet, lidar->data_src);
  uint32_t remaining_time = timestamp % publish_period_ns_;
  uint32_t diff_time = publish_period_ns_ - remaining_time;
  /** Get start time, down to the period boundary */
  if (diff_time > (publish_period_ns_ / 4)) {
    // ROS_INFO("0 : %u", diff_time);
    *start_time = timestamp - remaining_time;
    return 0;
  } else if (diff_time <= lidar->packet_interval_max) {
    *start_time = timestamp;
    return 0;
  } else {
    /** Skip some packets up to the period boundary*/
    // ROS_INFO("2 : %u", diff_time);
    do {
      if (QueueIsEmpty(queue)) {
        break;
      }
      QueuePopUpdate(queue); /* skip packet */
      QueuePrePop(queue, storage_packet);
      uint32_t last_remaning_time = remaining_time;
      timestamp = GetStoragePacketTimestamp(storage_packet, lidar->data_src);
      remaining_time = timestamp % publish_period_ns_;
      /** Flip to another period */
      if (last_remaning_time > remaining_time) {
        // ROS_INFO("Flip to another period, exit");
        break;
      }
      diff_time = publish_period_ns_ - remaining_time;
    } while (diff_time > lidar->packet_interval);

    /* the remaning packets in queue maybe not enough after skip */
    return -1;
  }
}




void Lddc::FillPointsToCustomMsg(TimeCloud& tCloud, \
    LivoxPointXyzrtl* src_point, uint32_t num, uint32_t offset_time, \
    uint32_t point_interval, uint32_t echo_num) {
  LivoxPointXyzrtl* point_xyzrtl = (LivoxPointXyzrtl*)src_point;
  for (uint32_t i = 0; i < num; i++) {
    PointType point;
    int64_t tm=0;
    if (echo_num > 1) { /** dual return mode */
      tm = (offset_time + (i / echo_num) * point_interval);
    } else {
      tm = (offset_time + i * point_interval);
    }
    point.x = point_xyzrtl->x;
    point.y = point_xyzrtl->y;
    point.z = point_xyzrtl->z;
    point.normal_x=double(tm)/1000000000.0;
    point.intensity = point_xyzrtl->reflectivity;
    //point.tag = point_xyzrtl->tag;
    point.normal_y = point_xyzrtl->line;
    ++point_xyzrtl;
    tCloud.cloud->points.push_back(point);
  }
}

uint32_t Lddc::PublishCustomPointcloud(LidarDataQueue *queue,
                                       uint32_t packet_num, uint8_t handle)
{
  static uint32_t msg_seq = 0;
  uint64_t timestamp = 0;
  uint64_t last_timestamp = 0;

  StoragePacket storage_packet;
  LidarDevice *lidar = &lds_->lidars_[handle];
  if (GetPublishStartTime(lidar, queue, &last_timestamp, &storage_packet))
  {
    /* the remaning packets in queue maybe not enough after skip */
    return 0;
  }

  TimeCloud Tcloud;
  Tcloud.cloud.reset(new PointCloud);

  uint8_t point_buf[2048];
  uint8_t data_source = lds_->lidars_[handle].data_src;
  uint32_t line_num = GetLaserLineNumber(lidar->info.type);
  uint32_t echo_num = GetEchoNumPerPoint(lidar->raw_data_type);
  uint32_t point_interval = GetPointInterval(lidar->info.type);
  uint32_t published_packet = 0;
  uint32_t packet_offset_time = 0;  /** uint:ns */
  uint32_t is_zero_packet = 0;
  while (published_packet < packet_num)
  {
    QueuePrePop(queue, &storage_packet);
    LivoxEthPacket *raw_packet =
            reinterpret_cast<LivoxEthPacket *>(storage_packet.raw_data);
    timestamp = GetStoragePacketTimestamp(&storage_packet, data_source);
    int64_t packet_gap = timestamp - last_timestamp;
    if ((packet_gap > lidar->packet_interval_max) &&
            lidar->data_is_pubulished)
    {
      // ROS_INFO("Lidar[%d] packet time interval is %ldns", handle,
      // packet_gap);
      if (kSourceLvxFile != data_source)
      {
        timestamp = last_timestamp + lidar->packet_interval;
        ZeroPointDataOfStoragePacket(&storage_packet);
        is_zero_packet = 1;
      }
    }
    /** first packet */
    if (!published_packet)
    {
      packet_offset_time = 0;
      /** convert to ros time stamp */
      Tcloud.timebase = timestamp;
    }
    else
    {
      packet_offset_time = (uint32_t) (timestamp - Tcloud.timebase);
    }
    uint32_t single_point_num = storage_packet.point_num * echo_num;

    if (kSourceLvxFile != data_source)
    {
      PointConvertHandler pf_point_convert =
              GetConvertHandler(lidar->raw_data_type);
      if (pf_point_convert)
      {
        pf_point_convert(point_buf, raw_packet, lidar->extrinsic_parameter, \
            line_num);
      }
      else
      {
        /* Skip the packet */
        printf("Lidar[%d] unkown packet type[%d]\n", handle,
               lidar->raw_data_type);
        break;
      }
    }
    else
    {
      LivoxPointToPxyzrtl(point_buf, raw_packet, lidar->extrinsic_parameter, \
          line_num);
    }
    LivoxPointXyzrtl *dst_point = (LivoxPointXyzrtl *) point_buf;
    FillPointsToCustomMsg(Tcloud, dst_point, single_point_num, \
        packet_offset_time, point_interval, echo_num);

    if (!is_zero_packet)
    {
      QueuePopUpdate(queue);
    }
    else
    {
      is_zero_packet = 0;
    }

    last_timestamp = timestamp;
    ++published_packet;
  }

  Tcloud.timebase /= 1000000000.0;
  if (private_pub_ != nullptr)
  {
    private_pub_(Tcloud, handle);
  }

  if (!lidar->data_is_pubulished)
  {
    lidar->data_is_pubulished = true;
  }
  return published_packet;
}

uint32_t Lddc::PublishImuData(LidarDataQueue *queue, uint32_t packet_num,
                              uint8_t handle) {
  uint64_t timestamp = 0;
  uint32_t published_packet = 0;

  /*sensor_msgs::Imu imu_data;
  imu_data.header.frame_id = "livox_frame";*/
  ImuData imu_data;

  uint8_t data_source = lds_->lidars_[handle].data_src;
  StoragePacket storage_packet;
  QueuePrePop(queue, &storage_packet);
  LivoxEthPacket *raw_packet =
      reinterpret_cast<LivoxEthPacket *>(storage_packet.raw_data);
  timestamp = GetStoragePacketTimestamp(&storage_packet, data_source);
  if (timestamp >= 0) {
    imu_data.timebase =timestamp / 1000000000.0;  // to ros time stamp
  }

  uint8_t point_buf[2048];
  LivoxImuDataProcess(point_buf, raw_packet);

  LivoxImuPoint *imu = (LivoxImuPoint *)point_buf;
  imu_data.angular_velocity.x = imu->gyro_x;
  imu_data.angular_velocity.y = imu->gyro_y;
  imu_data.angular_velocity.z = imu->gyro_z;
  imu_data.linear_acceleration.x = imu->acc_x;
  imu_data.linear_acceleration.y = imu->acc_y;
  imu_data.linear_acceleration.z = imu->acc_z;

  QueuePopUpdate(queue);
  ++published_packet;
  if (private_imu_pub_!= nullptr)
  {
    private_imu_pub_(imu_data,handle);
  }

  return published_packet;
}

int Lddc::RegisterLds(Lds *lds) {
  if (lds_ == nullptr) {
    lds_ = lds;
    return 0;
  } else {
    return -1;
  }
}

void Lddc::PollingLidarPointCloudData(uint8_t handle, LidarDevice *lidar)
{
  LidarDataQueue *p_queue = &lidar->data;
  if (p_queue->storage_packet == nullptr)
  {
    return;
  }

  while (!QueueIsEmpty(p_queue) && !exit_)
  {
    uint32_t used_size = QueueUsedSize(p_queue);
    uint32_t onetime_publish_packets = lidar->onetime_publish_packets;
    if (used_size < onetime_publish_packets)
    {
      break;
    }

    PublishCustomPointcloud(p_queue, onetime_publish_packets, handle);
  }
}

void Lddc::PollingLidarImuData(uint8_t handle, LidarDevice *lidar) {
  LidarDataQueue *p_queue = &lidar->imu_data;
  if (p_queue->storage_packet == nullptr) {
    return;
  }

  while (!QueueIsEmpty(p_queue)&&!exit_) {
    PublishImuData(p_queue, 1, handle);
  }
}

void Lddc::DistributeLidarData(void)
{
  if (lds_ == nullptr)
  {
    return;
  }
  lds_->semaphore_.Wait();
  for (uint32_t i = 0; i < lds_->lidar_count_; i++)
  {
    uint32_t lidar_id = i;
    LidarDevice *lidar = &lds_->lidars_[lidar_id];
    LidarDataQueue *p_queue = &lidar->data;
    if ((kConnectStateSampling != lidar->connect_state) ||
            (p_queue == nullptr))
    {
      continue;
    }
    PollingLidarPointCloudData(lidar_id, lidar);
    PollingLidarImuData(lidar_id, lidar);
  }

  if (lds_->IsRequestExit())
  {
    PrepareExit();
  }
}

void Lddc::Stop()
{
  if(capture_!= nullptr)
  {
    exit_=true;
    capture_->join();
    delete capture_;
    capture_= nullptr;
  }
}

void Lddc::Start()
{
  dataSourceType_=kSourceRawLidar;
  if (capture_== nullptr)
  {
    exit_=false;
    capture_=new std::thread(capture,this);
  }

}

void Lddc::Start(std::string lvx)
{
  dataSourceType_=kSourceLvxFile;
  lvx_file_=lvx;
  if (capture_== nullptr)
  {
    exit_=false;
    capture_=new std::thread(capture,this);
  }
}

void Lddc::capture(Lddc* p)
{
  int ret = 0;
  if (p->dataSourceType_ == kSourceRawLidar)
  {
    printf("Data Source is raw lidar.\n");
    std::string user_config_path;
    std::vector<std::string> bd_code_list;
    livox_ros::LdsLidar *read_lidar = livox_ros::LdsLidar::GetInstance(1000 / p->publish_frq_);
    p->RegisterLds(static_cast<livox_ros::Lds *>(read_lidar));
    ret = read_lidar->InitLdsLidar(bd_code_list, user_config_path.c_str());
    if (!ret)
    {
      printf("Init lds lidar success!\n");
    }
    else
    {
      perror("Init lds lidar fail!\n");
      return;
    }

  }
  else if (p->dataSourceType_ == kSourceLvxFile)
  {

    std::string lvx_file = p->lvx_file_;
    if (!IsFilePathValid(lvx_file.c_str()))
    {
      printf("File path invalid : %s !\n", lvx_file.c_str());
    }
    /* std::string rosbag_file_path;
     int path_end_pos = lvx_file.find_last_of('.');
     rosbag_file_path = lvx_file.substr(0, path_end_pos);
     rosbag_file_path += ".bag";*/

    LdsLvx *read_lvx = LdsLvx::GetInstance(1000 / p->publish_frq_);
    p->RegisterLds(static_cast<Lds *>(read_lvx));
    //p->CreateBagFile(rosbag_file_path);
    int ret = read_lvx->InitLdsLvx(lvx_file.c_str());
    if (!ret) {
      perror("Init lds lvx file success!\n");
    } else {
      perror("Init lds lvx file fail!\n");
    }
  }

  while (p->exit_ == false)
  {
    p->DistributeLidarData();
    std::this_thread::sleep_for(std::chrono::nanoseconds(10));
  }

}



void Lddc::PrepareExit(void) {
  if (lds_) {
    lds_->PrepareExit();
    lds_ = nullptr;
  }
}



}  // namespace livox_ros
