
#include "include/livox_ros_driver.h"

#include <chrono>
#include <vector>
#include <csignal>


#include "lddc.h"
#include "lds_hub.h"
#include "lds_lidar.h"
#include "lds_lvx.h"
#include "livox_sdk.h"

#include <pangolin/var/var.h>
#include <pangolin/var/varextra.h>
#include <pangolin/gl/gl.h>
#include <pangolin/gl/gldraw.h>
#include <pangolin/display/display.h>
#include <pangolin/display/view.h>
#include <pangolin/display/widgets.h>
#include <pangolin/display/default_font.h>
#include <pangolin/handler/handler.h>


using namespace livox_ros;

class LockCloud
{
 public:
    LockCloud()
    {
      cloud_.reset(new pcl::PointCloud<PointType>);
    }
    void Lock(){
      lock_.lock();
    }
    void unlock(){
      lock_.unlock();
    }
    pcl::PointCloud<PointType>::Ptr Get()
    {
      return cloud_;
    }
    void Set(pcl::PointCloud<PointType>::Ptr cloud)
    {
      lock_.lock();
      cloud_=cloud;
      lock_.unlock();
    }

 protected:
    std::mutex lock_;
    pcl::PointCloud<PointType>::Ptr cloud_;
};


LockCloud scan_cloud;
LockCloud localmap_cloud;
Eigen::Matrix4d cur_pose;
double pose_line_length=1.0;
//////////////////

const int32_t kSdkVersionMajorLimit = 2;


void CloudCallback(TimeCloud cloud,int handle)
{
  scan_cloud.Set(cloud.cloud);
  printf("time:%f lidar %d callback points:%d\n",cloud.timebase,handle,cloud.cloud->size());
}

void ImuDataCallback(ImuData imu,int handle)
{
  printf("time:%f lidar %d imu called\n",imu.timebase,handle);
}

int main(int argc, char **argv) {

  printf("Livox Ros Driver Version: %s\n", LIVOX_ROS_DRIVER_VERSION_STRING);

  /** Check sdk version */
  LivoxSdkVersion _sdkversion;
  GetLivoxSdkVersion(&_sdkversion);
  if (_sdkversion.major < kSdkVersionMajorLimit) {
    printf("The SDK version[%d.%d.%d] is too low\n", _sdkversion.major,
             _sdkversion.minor, _sdkversion.patch);
    return 0;
  }

  /** Init default system parameter */
  int xfer_format = kLivoxCustomMsg;   //1
  int multi_topic = 0;
  int data_src = kSourceRawLidar;
  double publish_freq  = 10.0; /* Hz */
  int output_type      = kOutputToRos;
  std::string frame_id = "livox_frame";
  bool lidar_bag = true;
  bool imu_bag   = false;

  /*

  livox_node.getParam("publish_freq", publish_freq);
  livox_node.getParam("enable_lidar_bag", lidar_bag);
  livox_node.getParam("enable_imu_bag", imu_bag);
   */
  if (publish_freq > 100.0) {
    publish_freq = 100.0;
  } else if (publish_freq < 0.1) {
    publish_freq = 0.1;
  } else {
    publish_freq = publish_freq;
  }

  /** Lidar data distribute control and lidar data source set */
  Lddc *lddc = new Lddc(publish_freq);


  lddc->SetCloudCallback(CloudCallback);
  lddc->SetImuCallback(ImuDataCallback);
  int ret = 0;
  if (data_src == kSourceRawLidar)
  {
    printf("Data Source is raw lidar.\n");
    std::string user_config_path;
    std::vector<std::string> bd_code_list;
    LdsLidar *read_lidar = LdsLidar::GetInstance(1000 / publish_freq);
    lddc->RegisterLds(static_cast<Lds *>(read_lidar));
    ret = read_lidar->InitLdsLidar(bd_code_list, user_config_path.c_str());
    if (!ret) {
      printf("Init lds lidar success!\n");
    } else {
      perror("Init lds lidar fail!\n");
    }
  }


  pangolin::CreateWindowAndBind("Main", 1280, 960);

  // 3D Mouse handler requires depth testing to be enabled
  glEnable(GL_DEPTH_TEST);

  // Define Camera Render Object (for view / scene browsing)
  pangolin::OpenGlRenderState s_cam(
          pangolin::ProjectionMatrix(1280, 960, 840, 840, 640, 480, 0.1, 1000),
          pangolin::ModelViewLookAt(-5, 0, 0, 5, 0, 0, pangolin::AxisZ)
  );

  // Choose a sensible left UI Panel width based on the width of 20
  // charectors from the default font.
  const int UI_WIDTH = 20 * pangolin::default_font().MaxWidth();

  // Add named OpenGL viewport to window and provide 3D Handler
  pangolin::View &d_cam = pangolin::CreateDisplay()
          .SetBounds(0.0, 1.0, pangolin::Attach::Pix(UI_WIDTH), 1.0, 1280.0f / 960.0f)
          .SetHandler(new pangolin::Handler3D(s_cam));

  // Add named Panel and bind to variables beginning 'ui'
  // A Panel is just a View with a default layout and input handling
  pangolin::CreatePanel("ui")
          .SetBounds(0.0, 1.0, 0.0, pangolin::Attach::Pix(UI_WIDTH));

  // Safe and efficient binding of named variables.
  // Specialisations mean no conversions take place for exact types
  // and conversions between scalar types are cheap.
  pangolin::Var<bool> a_button("ui.A_Button", false, false);
  pangolin::Var<double> a_double("ui.A_Double", 1, 0.33, 50);
  pangolin::Var<int> an_int("ui.An_Int", 2, 0, 5);
  pangolin::Var<double> a_double_log("ui.Log_scale", 3, 1, 1E4, true);
  pangolin::Var<bool> a_checkbox("ui.A_Checkbox", false, true);
  pangolin::Var<int> an_int_no_input("ui.An_Int_No_Input", 2);
  pangolin::Var<std::string> a_string("ui.A_String", "Edit ME!");

  // std::function objects can be used for Var's too. These work great with C++11 closures.
  pangolin::Var<std::function<void(void)>> save_window("ui.Save_Window", []() {
    pangolin::SaveWindowOnRender("window");
  });

  pangolin::Var<std::function<void(void)>> save_cube_view("ui.Save_Cube", [&d_cam]() {
    pangolin::SaveWindowOnRender("cube", d_cam.v);
  });

  // Demonstration of how we can register a keyboard hook to alter a Var
  pangolin::RegisterKeyPressCallback(pangolin::PANGO_CTRL + 'b', [&]() {
    //a_double = 3.5;

  });

  // Default hooks for exiting (Esc) and fullscreen (tab).

  glEnable(GL_BLEND);   // 启用混合
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  double zoom = 1.0;
  while (!pangolin::ShouldQuit())
  {
    // Clear entire screen
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if (pangolin::Pushed(a_button))
      std::cout << "You Pushed a button!" << std::endl;

    if (a_double.GuiChanged())
    {
      pose_line_length=a_double.Get();
      if (pose_line_length<1)
        pose_line_length=1;
      lddc->SetPublishFrq(pose_line_length);
    }

    a_string="123654";

    // Overloading of Var<T> operators allows us to treat them like
    // their wrapped types, eg:
    if (a_checkbox)
      an_int = (int) a_double;

    an_int_no_input = an_int;

    if (d_cam.IsShown())
    {
      // Activate efficiently by object
      //s_cam.SetModelViewMatrix(pangolin::ModelViewLookAt(odom_pt[0],odom_pt[1],60, odom_pt[0],odom_pt[1],0, pangolin::AxisY));

      d_cam.Activate(s_cam);
      pangolin::glDrawAxis(3);//三维坐标轴，红——x轴，绿——y轴，蓝——z轴


      glBegin(GL_POINTS);
      glColor3f(1, 1, 1);
      glPointSize(0.5);
      scan_cloud.Lock();
      for (int i = 0; i < scan_cloud.Get()->size(); ++i)
      {
        PointType pt = scan_cloud.Get()->points[i];
        glVertex3d(pt.x * zoom, pt.y * zoom, pt.z * zoom);
      }
      scan_cloud.unlock();

      glEnd();

    }

    // Swap frames and Process Events
    lddc->DistributeLidarData();
    pangolin::FinishFrame();
    //std::this_thread::sleep_for(std::chrono::microseconds(1));

  }

  return 0;
}
